/* Search Query Parser
 * Version 1.1
 * Last change: 29. January 2006
 * 
 * Converted to C# and extended by Thomas Bandt (http://blog.thomasbandt.de)
 * Based on http://www.glengamoi.com/pipermail/asp.net/2005-May/015598.html
 * 
 * For further informations see: http://blog.thomasbandt.de
 */

using System;
using System.Collections.Generic;
using System.Text;

namespace SnippetCenter.Bll.Search
{

    /// <summary>
    /// Defines the default operator to combine the several conditions. At
    /// the moment there are no other operators than OR and AND which makes
    /// sense.
    /// </summary>
    public enum QueryOperator
    {
        OR,
        AND
    }

    /// <summary>
    /// What type of query should be returned - one which works with Like %%
    /// or one which uses the DIFFERENCE().
    /// </summary>
    public enum QueryType
    {
        Like,
        Difference
    }

    /// <summary>
    /// This class offers in fact the Parse method which parses a query to
    /// a usable sql where condition statement
    /// </summary>
    public class QueryParser
    {

        /// <summary>
        /// Contains the serveral terms for the internal parsing process
        /// </summary>
        private List<string> SearchTerms;

        /// <summary>
        /// Defines the characters which are parsed out the search string
        /// </summary>
        private const string ExcludedCharacters = "AND OR NOT ( )";

        /// <summary>
        /// Default Operator
        /// </summary>
        private QueryOperator defaultOperator = QueryOperator.OR;
        public QueryOperator DefaultOperator
        {
            get { return defaultOperator; }
            set { defaultOperator = value; }
        }

        /// <summary>
        /// Query Type
        /// </summary>
        private QueryType queryType = QueryType.Like;
        public QueryType QueryType
        {
            get { return queryType; }
            set { queryType = value; }
        }

        /// <summary>
        /// Represents the maximum value of difference allowed by
        /// using the DIFFERENCE() function
        /// </summary>
        private int minDifference = 3;
        public int MinDifference
        {
            get { return minDifference; }
            set { minDifference = value; }
        }

        /// <summary>
        /// Constructor
        /// </summary>
        public QueryParser()
        {
            SearchTerms = new List<string>();
        }

        /// <summary>
        /// Parsing procedure
        /// </summary>
        /// <param name="searchString">Contains the user input, for example "Peace NOT War"</param>
        /// <param name="fields">String Array, defines the fields in which should be looked for in the database</param>
        /// <returns>Returns what has to be right to SELECT * FROM Table WHERE </returns>
        public string Parse(string searchString, string[] fields)
        {
            if (!string.IsNullOrEmpty(searchString))
            {

                searchString += " ";
                bool inString = false;
                string tok = string.Empty;
                
                // Loop trough searchString and strip it
                for (int i = 0; i < searchString.Length; i++)
                {
                    char character = searchString[i];
                    switch (character)
                    {
                        case '"':
                            inString = !inString;
                            if (tok.Length > 0)
                            {
                                AddToSearchTerms(tok, !inString);
                                tok = string.Empty;
                            }
                            break;
                        case '(':
                            if (!inString)
                            {
                                if (tok.Length > 0)
                                {
                                    AddToSearchTerms(tok, false);
                                    tok = string.Empty;
                                }
                                AddToSearchTerms(character.ToString(), false);
                            }
                            else
                            {
                                tok += character.ToString();
                            }
                            break;
                        case ')':
                            goto case '(';
                        default:
                            if (inString || !character.Equals(' '))
                            {
                                tok += character.ToString();
                            }
                            else
                            {
                                if (tok.Length > 0)
                                {
                                    AddToSearchTerms(tok, false);
                                    tok = string.Empty;
                                }
                            }
                            break;
                    }
                }

                StringBuilder where = new StringBuilder(string.Empty);
                bool wasString = false;
                string defaultOperator = DefaultOperator.ToString();

                // Loop trough all SearchTerms items
                for (int i = 0; i < SearchTerms.Count; i++)
                {
                    if (ExcludedCharacters.IndexOf(SearchTerms[i]) > -1)
                    {
                        // Prevent constructs like this one: WHERE AND NOT ...
                        if (i > 0 || SearchTerms[i].ToUpper() != "AND")
                        {
                            // Add operator to statement
                            where.AppendFormat(" {0} ", SearchTerms[i]);
                            wasString = false;
                        }
                    }
                    else
                    {
                        // Add the default operator to link the single statements
                        if (wasString) where.AppendFormat(" {0} ", defaultOperator);
                        // Add the term for each database column (field).
                        switch (QueryType)
                        {
                            case QueryType.Difference:
                                where.Append("(");
                                for (int y = 0; y < fields.Length; y++)
                                {
                                    where.AppendFormat("DIFFERENCE({0},{1}) >= {2}", fields[y], SearchTerms[i], MinDifference);
                                    if (y + 1 < fields.Length)
                                        where.Append(" OR ");
                                }
                                where.Append(")");
                                break;
                            default:
                                where.AppendFormat(" ({0} LIKE {1}) ", string.Join(string.Format(" LIKE {0} OR ", SearchTerms[i]), fields), SearchTerms[i]);
                                break;
                        }
                        wasString = true;
                    }
                }

                // Return the Where statement. Ready :-)
                return where.ToString();

            }
            else
            {
                // When searchString is empty, then return an empty string.
                return string.Empty;
            }

        }

        /// <summary>
        /// Buffers an operator or a term to SearchTerms list
        /// </summary>
        /// <param name="term">Contains the string which has to be added.</param>
        /// <param name="isTerm">Define whether the string is a term or an operator.</param>
        private void AddToSearchTerms(string term, bool isTerm)
        {
            // Add an operator
            if (!isTerm && ExcludedCharacters.ToUpper().IndexOf(term.ToUpper()) > -1)
            { 
                SearchTerms.Add(term.ToUpper()); 
            }
            // Add a new term
            else
            {
                switch (QueryType)
                {
                    case QueryType.Difference:
                        SearchTerms.Add(string.Format("'{0}'", term.Replace("'", "''")));
                        break;
                    default:
                        SearchTerms.Add(string.Format("'%{0}%'", term.Replace("'", "''")));
                        break;
                }
            }
        }

    }
}

//Original VBScript-Code by Claudius Ceteras postet at:
//http://www.glengamoi.com/pipermail/asp.net/2005-May/015598.html

//function parse(ByVal searchString,fields,defaultOperator)
//  dim tokens,i,inString,char,tok,where,wasStr
//  searchString = searchString & " "
//  tokens = Array()
//  inString=false
//  tok = ""
//  for i = 1 to len(searchString)
//    char = mid(searchString,i,1)

//    select case char
//    case """"
//      inString = not inString
//      if (tok<>"") then tokens=arrayAdd(tokens,tok,not inString) : tok = ""
//    case "(",")"
//        if not inString then
//          if (tok<>"") then tokens=arrayAdd(tokens,tok,false) : tok = ""
//          tokens = arrayAdd(tokens,char,false)
//        else
//          tok = tok & char
//        end if
//    case else
//          if inString or char<>" " then
//            tok = tok & char
//          else
//            if tok<>"" then tokens=arrayAdd(tokens,tok,false) : tok = ""
//          end if
//    end select
//  next

//  where="" : wasStr = false
//  for each tok in tokens
//    select case tok
//      case "(",")","AND","OR","NOT"
//        where = where & " " & tok & " "
//        wasStr = false
//      case else
//        if wasStr then where = where & " " & defaultOperator & " "

//        where = where & " (" & join(fields," LIKE " & tok & " OR ") & " LIKE " & tok & ") "

//        wasStr = true
//    end select
//  next
//  parse = where
//end function

//function arrayAdd(arr,ByVal str,isString)
//  dim newArr()
//  dim i
//  redim newArr(ubound(arr)+1)
//  for i = 0 to ubound(arr)
//    newArr(i) = arr(i)
//  next

//  if not isString and (ucase(str)="AND" or ucase(str)="OR" _
//          or ucase(str)="NOT" or str="(" or str=")") then
//    str = ucase(str)
//  else
//    str = "'%" & replace(str,"'","''") & "%'"
//  end if

//  newArr(ubound(newArr)) = str
//  arrayAdd = newArr
//end function